@@ -1,7 +1,6 @@
 /// # Operators on array and shapes
 mod broadcast;
 pub(crate) mod concat;
-mod constant_of_shape;
 mod gather;
 mod gather_elements;
 mod gather_nd;
@@ -15,7 +14,6 @@ mod tile;
 
 pub use self::broadcast::MultiBroadcastTo;
 pub use self::concat::{ConcatSlice, TypedConcat};
-pub use self::constant_of_shape::ConstantOfShape;
 pub use self::gather::Gather;
 pub use self::gather_elements::GatherElements;
 pub use self::gather_nd::GatherNd;
