@@ -298,6 +298,28 @@ impl Scan {
         Ok(None)
     }
 
+    fn declutter_discard_empty_output_mapping_with_body_output(
+        &self,
+        model: &TypedModel,
+        node: &TypedNode,
+    ) -> TractResult<Option<TypedModelPatch>> {
+        for (ix, om) in self.output_mapping.iter().enumerate() {
+            if om.last_value_slot.is_none() && om.full_slot.is_none() && !om.state {
+                let mut new_op = self.clone();
+                new_op.output_mapping.remove(ix);
+                new_op.body.outputs.remove(ix);
+                new_op.decluttered = false;
+                return Ok(Some(TypedModelPatch::replace_single_op(
+                    model,
+                    node,
+                    &node.inputs,
+                    new_op,
+                )?));
+            }
+        }
+        Ok(None)
+    }
+
     fn declutter_pull_batcheable_input(
         &self,
         model: &TypedModel,
@@ -384,12 +406,26 @@ impl Scan {
         node: &TypedNode,
     ) -> TractResult<Option<TypedModelPatch>> {
         for (model_ix, mapping) in self.output_mapping.iter().enumerate() {
-            let slot = if let Some(slot) = mapping.full_slot { slot } else { continue };
+            let mut is_used = false;
+            for slot in &[mapping.full_slot, mapping.last_value_slot] {
+                if let Some(slot) = slot {
+                    if node.outputs[*slot].successors.len() > 0
+                        || model.outputs.contains(&(node.id, *slot).into())
+                    {
+                        is_used = true;
+                    }
+                }
+            }
+            if !is_used {
+                continue;
+            }
             let emitter_outlet = self.body.output_outlets()?[model_ix];
             let emitter_node = self.body.node(emitter_outlet.node);
-            if emitter_node.outputs[emitter_outlet.slot].successors.len() > 1
-                || emitter_node.inputs.len() > 1
+            if emitter_node.outputs[emitter_outlet.slot].successors.len() > 0
+                || mapping.state
+                || mapping.chunk > 1
             {
+                // continue if both last_value and full values are exported
                 continue;
             }
             let invariants = emitter_node.op.invariants(&self.body, &emitter_node)?;
@@ -398,10 +434,42 @@ impl Scan {
             } else {
                 continue;
             };
+
             let mut fixed_body = self.body.clone();
-            fixed_body.outputs[model_ix] = emitter_node.inputs[0];
             let mut output_mapping = self.output_mapping.clone();
-            output_mapping[model_ix] = OutputMapping { axis: axis_before, ..mapping.clone() };
+
+            let mut full_values_slots = vec![];
+            let mut last_values_slots = vec![];
+            let mut current_output_number = node.outputs.len();
+            for input in &emitter_node.inputs {
+                if let Some(position) = fixed_body.outputs.iter().position(|o| o == input) {
+                    let mut mapping = &mut output_mapping[position];
+                    if mapping.last_value_slot.is_none() {
+                        mapping.last_value_slot = Some(current_output_number);
+                        current_output_number += 1;
+                    }
+                    if mapping.full_slot.is_none() {
+                        mapping.axis = axis_before;
+                        mapping.full_slot = Some(current_output_number);
+                        current_output_number += 1;
+                    }
+                    full_values_slots.push(mapping.full_slot.unwrap());
+                    last_values_slots.push(mapping.last_value_slot.unwrap());
+                } else {
+                    fixed_body.outputs.push(*input);
+                    let full_slot = Some(current_output_number);
+                    let last_value_slot = Some(current_output_number + 1);
+                    current_output_number += 2;
+                    output_mapping.push(OutputMapping {
+                        axis: axis_before,
+                        full_slot,
+                        last_value_slot,
+                        ..mapping.clone()
+                    });
+                    full_values_slots.push(full_slot.unwrap());
+                    last_values_slots.push(last_value_slot.unwrap());
+                }
+            }
 
             let mut outside_patch = TypedModelPatch::default();
             let inputs = node
@@ -418,19 +486,18 @@ impl Scan {
                 seq_length_input_slot: self.seq_length_input_slot,
             };
             let scan_outputs = outside_patch.wire_node(&*node.name, new_op, &*inputs)?;
-            let wire = outside_patch.wire_node(
-                &*emitter_node.name,
-                emitter_node.op.clone(),
-                &[scan_outputs[slot]],
-            )?[0];
-            for ix in 0..node.outputs.len() {
-                if ix == slot {
-                    outside_patch.shunt_outside(model, OutletId::new(node.id, ix), wire)?;
-                } else {
+            let output = mapping.full_slot.unwrap();
+            let inputs =
+                full_values_slots.iter().map(|slot| scan_outputs[*slot]).collect::<TVec<_>>();
+            let wire =
+                outside_patch.wire_node(&*emitter_node.name, emitter_node.op.clone(), &inputs)?[0];
+            outside_patch.shunt_outside(model, OutletId::new(node.id, output), wire)?;
+            for output_slot in 0..node.outputs.len() {
+                if output_slot != output {
                     outside_patch.shunt_outside(
                         model,
-                        OutletId::new(node.id, ix),
-                        scan_outputs[ix],
+                        OutletId::new(node.id, output_slot),
+                        OutletId::new(scan_outputs[0].node, output_slot),
                     )?;
                 }
             }
@@ -617,7 +684,7 @@ impl TypedOp for Scan {
             if let Some(slot) = output.full_slot {
                 let mut shape = fact.shape.clone();
                 let scanning_dim =
-                    output.full_dim_hint.clone().unwrap_or(shape[output.axis].maybe_mul(&iters)?);
+                    output.full_dim_hint.clone().unwrap_or(shape[output.axis].clone() * &iters);
                 shape.set(output.axis, scanning_dim);
                 outputs.push((slot, TypedFact::dt_shape(fact.datum_type, shape)));
             }
@@ -707,6 +774,7 @@ impl TypedOp for Scan {
                 {
                     trace!(stringify!($func));
                     r.push_context(stringify!($func));
+                    // FIXME: wtf ?
                     for node in r.nodes() {
                         if let Some(scan) = node.op_as::<Self>() {
                             scan.body.invariants().unwrap();
@@ -723,6 +791,7 @@ impl TypedOp for Scan {
         pass!(declutter_pull_batcheable_output);
         pass!(declutter_const_initializer);
         pass!(declutter_discard_useless_outer_output);
+        pass!(declutter_discard_empty_output_mapping_with_body_output);
         Ok(None)
     }
 
@@ -741,6 +810,7 @@ impl TypedOp for Scan {
                 .iter()
                 .map(|om| om.concretize_dims(values))
                 .collect::<TractResult<Vec<_>>>()?,
+            body: self.body.concretize_dims(values)?,
             ..self.clone()
         };
         target.wire_node(&node.name, op, &inputs)
